const Chunk = @import("./kym/Chunk.zig");

const Table = @import("./kym/Table.zig");

const coral = @import("coral");

const file = @import("./file.zig");

const gfx = @import("./gfx.zig");

const tokens = @import("./kym/tokens.zig");

const tree = @import("./kym/tree.zig");

///
/// Fixed-length integer number type.
///
pub const Fixed = i32;

///
/// Floating point real number type.
///
pub const Float = f64;

///
/// Supertype for all numeric types.
///
pub const Numeric = union (enum) {
	fixed: Fixed,
	float: Float,
	vector2: Vector2,
	vector3: Vector3,
};

///
/// Runtime environment virtual machine state.
///
/// The environment operates on a mostly stack-based architecture, with nearly all functions "returning" computations by
/// writing them to the local values stack within. Functions may be chained together to perform sequential mutations on
/// data and either retrieved or discarded once the computation is done via [RuntimeEnv.pop] or [RuntimeEnv.discard]
/// respectively.
///
/// *Note* that improperly cleaning up data pushed to the stack can result in accidental stack overflows, however also
/// note that all locals are always cleaned when exiting their call frame.
///
pub const RuntimeEnv = struct {
	allocator: coral.io.Allocator,
	options: Options,
	interned_symbols: SymbolSet,
	locals: LocalList,
	frames: FrameList,

	const FrameList = coral.list.Stack(struct {
		callable: *RuntimeObj,
		locals_top: usize,
		arg_count: u8,

		const Self = @This();

		fn args_of(self: Self, env: *RuntimeEnv) []?*RuntimeObj {
			return env.locals.values[self.locals_top .. (self.locals_top + self.arg_count)];
		}

		fn locals_of(self: Self, env: *RuntimeEnv) []?*RuntimeObj {
			return env.locals.values[self.locals_top ..];
		}
	});

	const LocalList = coral.list.Stack(?*RuntimeObj);

	///
	/// Optional settings for a [RuntimeEnv].
	///
	pub const Options = struct {
		import_access: file.Access = .null,
		print: ?*const Printer = null,
		print_error: ?*const Printer = null,

		///
		/// Byte string printing function.
		///
		pub const Printer = fn ([]const coral.io.Byte) void;
	};

	const SymbolSet = coral.map.StringTable([:0]coral.io.Byte);

	///
	/// Attempts to push the argument located at `arg_index` to the top of `self`, pushing `null` instead if the given
	/// argument does not exist or was provided as a nil value.
	///
	/// Arguments are indexed according to the order they are passed, with `0` referring to the first argument position.
	///
	/// A [RuntimeError] is returned if `self` is out of memory or the virtual machine is not inside a managed call
	/// frame.
	///
	/// `self` is returned for function chaining.
	///
	pub fn arg_get(self: *RuntimeEnv, arg_index: Local) RuntimeError!*RuntimeEnv {
		const frame = self.frames.peek() orelse {
			return self.raise(error.IllegalState, "cannot get args outside of a call frame", .{});
		};

		const args = frame.args_of(self);

		if (arg_index < args.len) {
			return self.push(args[arg_index]);
		}

		try self.locals.push_one(null);

		return self;
	}

	///
	/// Attempts to pop the top-most value of `self` and call it with `local_arg_count` as the number of locals prior to
	/// it in `self` that are intended to be arguments to it. Once the callable returns, the locals marked as arguments
	/// are popped as well.
	///
	/// A `local_arg_count` of `0` will call the function with no arguments and pop nothing other than the callable from
	/// `self` during invocation.
	///
	/// A [RuntimeError] is returned if `self` is out of memory, the top-most local is nil or not callable, or the
	/// callable raises a runtime error during invocation.
	///
	/// `self` is returned for function chaining.
	///
	pub fn call(self: *RuntimeEnv, local_arg_count: Local) RuntimeError!*RuntimeEnv {
		const callable = try self.expect_object(self.pop());

		defer self.release(callable);

		const result = get_result: {
			try self.frames.push_one(.{
				.locals_top = self.locals.values.len - local_arg_count,
				.callable = callable.internal().acquire(),
				.arg_count = local_arg_count,
			});

			defer {
				const popped_frame = self.frames.pop().?;

				self.release(popped_frame.callable);
				coral.debug.assert(popped_frame.locals_top < self.locals.values.len);

				var to_discard = self.locals.values.len - popped_frame.locals_top;

				while (to_discard != 0) : (to_discard -= 1) {
					self.discard();
				}
			}

			const frame = &self.frames.values[self.frames.values.len - 1];

			break: get_result try switch (frame.callable.internal().payload) {
				.syscall => |syscall| syscall(self),

				.dynamic => |dynamic| dynamic.typeinfo().call(.{
					.userdata = dynamic.userdata(),
					.env = self,
				}),

				else => self.raise(error.TypeMismatch, "{typename} is not callable", .{
					.typename = frame.callable.typename(),
				}),
			};
		};

		errdefer {
			if (result) |object| {
				self.release(object);
			}
		}

		try self.locals.push_one(result);

		return self;
	}

	///
	/// Attempts to pop and compare the top-most local in `self` with `rhs_comparable`, pushing a true value if the
	/// local is greater than `rhs_comparable`, otherwise pushing a false value.
	///
	/// A [RuntimeError] is returned if `self` is out of memory, the top-most local is nil or not comparable, or
	/// `rhs_comparable` is not comparable.
	///
	/// `self` is returned for function chaining.
	///
	pub fn compare_greater(self: *RuntimeEnv, rhs_comparable: *RuntimeObj) RuntimeError!*RuntimeEnv {
		const lhs_comparable = try self.expect_object(self.pop());

		defer self.release(lhs_comparable);

		return switch (lhs_comparable.internal().payload) {
			.fixed => |lhs_fixed| switch (rhs_comparable.internal().payload) {
				.fixed => |rhs_fixed| self.new_boolean(@as(Float, @floatFromInt(lhs_fixed)) > @as(Float, @floatFromInt(rhs_fixed))),
				.float => |rhs_float| self.new_boolean(@as(Float, @floatFromInt(lhs_fixed)) > rhs_float),

				else => self.raise(error.TypeMismatch, "right-hand {typename} is not comparable", .{
					.typename = rhs_comparable.typename(),
				}),
			},

			.float => |lhs_float| switch (rhs_comparable.internal().payload) {
				.float => |rhs_float| self.new_boolean(lhs_float > rhs_float),
				.fixed => |rhs_fixed| self.new_boolean(lhs_float > @as(Float, @floatFromInt(rhs_fixed))),

				else => self.raise(error.TypeMismatch, "right-hand {typename} is not comparable", .{
					.typename = rhs_comparable.typename(),
				}),
			},

			else => self.raise(error.TypeMismatch, "left-hand {typename} is not comparable", .{
				.typename = lhs_comparable.typename(),
			}),
		};
	}

	///
	/// Attempts to pop and compare the top-most local in `self` with `rhs_comparable`, pushing a true value if the
	/// local is greater than or equal to `rhs_comparable`, otherwise pushing a false value.
	///
	/// A [RuntimeError] is returned if `self` is out of memory, the top-most local is nil or not comparable, or
	/// `rhs_comparable` is not comparable.
	///
	/// `self` is returned for function chaining.
	///
	pub fn compare_greater_equals(self: *RuntimeEnv, rhs_comparable: *RuntimeObj) RuntimeError!*RuntimeEnv {
		const lhs_comparable = try self.expect_object(self.pop());

		defer self.release(lhs_comparable);

		return switch (lhs_comparable.internal().payload) {
			.fixed => |lhs_fixed| switch (rhs_comparable.internal().payload) {
				.fixed => |rhs_fixed| self.new_boolean(@as(Float, @floatFromInt(lhs_fixed)) >= @as(Float, @floatFromInt(rhs_fixed))),
				.float => |rhs_float| self.new_boolean(@as(Float, @floatFromInt(lhs_fixed)) >= rhs_float),

				else => self.raise(error.TypeMismatch, "right-hand {typename} is not comparable", .{
					.typename = rhs_comparable.typename(),
				}),
			},

			.float => |lhs_float| switch (rhs_comparable.internal().payload) {
				.float => |rhs_float| self.new_boolean(lhs_float >= rhs_float),
				.fixed => |rhs_fixed| self.new_boolean(lhs_float >= @as(Float, @floatFromInt(rhs_fixed))),

				else => self.raise(error.TypeMismatch, "right-hand {typename} is not comparable", .{
					.typename = rhs_comparable.typename(),
				}),
			},

			else => self.raise(error.TypeMismatch, "left-hand {typename} is not comparable", .{
				.typename = lhs_comparable.typename(),
			}),
		};
	}

	///
	/// Attempts to pop and compare the top-most local in `self` with `rhs_comparable`, pushing a true value if the
	/// local is less than `rhs_comparable`, otherwise pushing a false value.
	///
	/// A [RuntimeError] is returned if `self` is out of memory, the top-most local is nil or not comparable, or
	/// `rhs_comparable` is not comparable.
	///
	/// `self` is returned for function chaining.
	///
	pub fn compare_less(self: *RuntimeEnv, rhs_comparable: *RuntimeObj) RuntimeError!*RuntimeEnv {
		const lhs_comparable = try self.expect_object(self.pop());

		defer self.release(lhs_comparable);

		return switch (lhs_comparable.internal().payload) {
			.fixed => |lhs_fixed| switch (rhs_comparable.internal().payload) {
				.fixed => |rhs_fixed| self.new_boolean(@as(Float, @floatFromInt(lhs_fixed)) < @as(Float, @floatFromInt(rhs_fixed))),
				.float => |rhs_float| self.new_boolean(@as(Float, @floatFromInt(lhs_fixed)) < rhs_float),

				else => return self.raise(error.TypeMismatch, "right-hand {typename} is not comparable", .{
					.typename = rhs_comparable.typename(),
				}),
			},

			.float => |lhs_float| switch (rhs_comparable.internal().payload) {
				.float => |rhs_float| self.new_boolean(lhs_float < rhs_float),
				.fixed => |rhs_fixed| self.new_boolean(lhs_float < @as(Float, @floatFromInt(rhs_fixed))),

				else => return self.raise(error.TypeMismatch, "right-hand {typename} is not comparable", .{
					.typename = rhs_comparable.typename(),
				}),
			},

			else => return self.raise(error.TypeMismatch, "left-hand {typename} is not comparable", .{
				.typename = lhs_comparable.typename(),
			}),
		};
	}

	///
	/// Attempts to pop and compare the top-most local in `self` with `rhs_comparable`, pushing a true value if the
	/// local is less than or equal to `rhs_comparable`, otherwise pushing a false value.
	///
	/// A [RuntimeError] is returned if `self` is out of memory, the top-most local is nil or not comparable, or
	/// `rhs_comparable` is not comparable.
	///
	/// `self` is returned for function chaining.
	///
	pub fn compare_less_equals(self: *RuntimeEnv, rhs_comparable: *RuntimeObj) RuntimeError!*RuntimeEnv {
		const lhs_comparable = try self.expect_object(self.pop());

		defer self.release(lhs_comparable);

		return switch (lhs_comparable.internal().payload) {
			.fixed => |lhs_fixed| switch (rhs_comparable.internal().payload) {
				.fixed => |rhs_fixed| self.new_boolean(@as(Float, @floatFromInt(lhs_fixed)) <= @as(Float, @floatFromInt(rhs_fixed))),
				.float => |rhs_float| self.new_boolean(@as(Float, @floatFromInt(lhs_fixed)) <= rhs_float),

				else => return self.raise(error.TypeMismatch, "right-hand {typename} is not comparable", .{
					.typename = rhs_comparable.typename(),
				}),
			},

			.float => |lhs_float| switch (rhs_comparable.internal().payload) {
				.float => |rhs_float| self.new_boolean(lhs_float <= rhs_float),
				.fixed => |rhs_fixed| self.new_boolean(lhs_float <= @as(Float, @floatFromInt(rhs_fixed))),

				else => return self.raise(error.TypeMismatch, "right-hand {typename} is not comparable", .{
					.typename = rhs_comparable.typename(),
				}),
			},

			else => return self.raise(error.TypeMismatch, "left-hand {typename} is not comparable", .{
				.typename = lhs_comparable.typename(),
			}),
		};
	}

	///
	/// Attempts to pop `local_concat_count` number of locals from the stack and concatenate them together as one long
	/// string object, pushing the result to the top of `self`.
	///
	/// A `local_concat_count` of `0` will push an empty string object and pop nothing from `self`.
	///
	/// A [RuntimeError] is returned if `self` is out of memory or one of the concatenated objects raises an error from
	/// within it's to_string implementation.
	///
	/// `self` is returned for function chaining.
	///
	pub fn concat(self: *RuntimeEnv, local_concat_count: Local) RuntimeError!*RuntimeEnv {
		if (local_concat_count == 0) {
			return self.new_string("");
		}

		const concated_buffer = build_buffer: {
			var concat_buffer = coral.list.ByteStack.init(self.allocator);

			errdefer concat_buffer.deinit();

			const concat_values = self.locals.values[(self.locals.values.len - local_concat_count) .. self.locals.values.len];

			for (concat_values) |concat_value| {
				const string = (try (try self.push(concat_value)).to_string()).pop().?;

				defer self.release(string);

				try concat_buffer.push_all(string.is_string().?);
			}

			{
				var concated_value_count = concat_values.len;

				while (concated_value_count != 0) : (concated_value_count -= 1) {
					self.discard();
				}
			}

			break: build_buffer try concat_buffer.pack();
		};

		errdefer self.allocator.deallocate(concated_buffer);

		const string = try RuntimeObj.allocate(self.allocator, .{
			.ref_count = 1,

			.payload = .{
				.string = .{
					.ptr = concated_buffer.ptr,
					.len = @intCast(concated_buffer.len),
				},
			},
		});

		errdefer self.release(string);

		try self.locals.push_one(string);

		return self;
	}

	///
	/// Deinitializes `self`, freeing all allocated virtual machine resources.
	///
	pub fn deinit(self: *RuntimeEnv) void {
		while (self.locals.pop()) |local| {
			if (local) |ref| {
				self.release(ref);
			}
		}

		{
			var iterable = self.interned_symbols.as_iterable();

			while (iterable.next()) |entry| {
				self.allocator.deallocate(entry.value);
			}
		}

		self.interned_symbols.deinit();
		self.locals.deinit();
		self.frames.deinit();
	}

	///
	/// Pops the top-most value of `self` and releases it.
	///
	pub fn discard(self: *RuntimeEnv) void {
		if (self.pop()) |popped| {
			self.release(popped);
		}
	}

	pub fn equals_nil(self: *RuntimeEnv) RuntimeError!*RuntimeEnv {
		if (self.pop()) |lhs_object| {
			defer self.release(lhs_object);

			return self.new_boolean(false);
		}

		return self.new_boolean(true);
	}

	pub fn equals_object(self: *RuntimeEnv, rhs_object: *RuntimeObj) RuntimeError!*RuntimeEnv {
		if (self.pop()) |lhs_object| {
			defer self.release(lhs_object);

			return self.new_boolean(lhs_object.equals(rhs_object));
		}

		return self.new_boolean(false);
	}

	pub fn expect_dynamic(self: *RuntimeEnv, value: *RuntimeObj, typeinfo: *const Typeinfo) RuntimeError![]coral.io.Byte {
		return value.is_dynamic(typeinfo) orelse self.raise(error.TypeMismatch, "expected dynamic object, not {typename}", .{
			.typename = value.typename(),
		});
	}

	pub fn expect_float(self: *RuntimeEnv, value: *RuntimeObj) RuntimeError!Float {
		return switch (value.internal().payload) {
			.fixed => |fixed| @floatFromInt(fixed),
			.float => |float| float,

			else => self.raise(error.TypeMismatch, "expected float type, not {typename}", .{
				.typename = value.typename(),
			}),
		};
	}

	pub fn expect_fixed(self: *RuntimeEnv, value: *RuntimeObj) RuntimeError!Fixed {
		return value.is_fixed() orelse self.raise(error.TypeMismatch, "expected fixed type, not {typename}", .{
			.typename = value.typename(),
		});
	}

	pub fn expect_numeric(self: *RuntimeEnv, value: *RuntimeObj) RuntimeError!Numeric {
		return value.is_numeric() orelse self.raise(error.TypeMismatch, "expected numeric type, not {typename}", .{
			.typename = value.typename(),
		});
	}

	pub fn expect_object(self: *RuntimeEnv, value: ?*RuntimeObj) RuntimeError!*RuntimeObj {
		return value orelse self.raise(error.TypeMismatch, "expected object type, not nil", .{});
	}

	pub fn expect_string(self: *RuntimeEnv, value: *RuntimeObj) RuntimeError![]const coral.io.Byte {
		return value.is_string() orelse self.raise(error.TypeMismatch, "expected string type, not {typename}", .{
			.typename = value.typename(),
		});
	}

	pub fn expect_symbol(self: *RuntimeEnv, value: *RuntimeObj) RuntimeError![*:0]const coral.io.Byte {
		return value.is_symbol() orelse self.raise(error.TypeMismatch, "expected symbol type, not {typename}", .{
			.typename = value.typename(),
		});
	}

	///
	/// Attempts to pop the top-most value in `self` and add `rhs_fixed` to it, pushing the result.
	///
	/// A [RuntimeError] is returned if `self` is out of memory or `rhs_fixed` is not addable with the top-most local.
	///
	/// `self` is returned for function chaining.
	///
	pub fn fixed_add(self: *RuntimeEnv, rhs_fixed: Fixed) RuntimeError!*RuntimeEnv {
		const addable = try self.expect_object(self.pop());

		defer self.release(addable);

		return switch (addable.internal().payload) {
			.fixed => |lhs_fixed| if (coral.math.checked_add(lhs_fixed, rhs_fixed)) |result|
					self.new_fixed(result)
				else
					self.new_float(@as(Float, @floatFromInt(lhs_fixed)) + @as(Float, @floatFromInt(rhs_fixed))),

			.float => |lhs_float| self.new_float(lhs_float + @as(Float, @floatFromInt(rhs_fixed))),
			.vector2 => |lhs_vector2| self.new_vector2(lhs_vector2.scalar_added(@floatFromInt(rhs_fixed))),
			.vector3 => |lhs_vector3| self.new_vector3(lhs_vector3.scalar_added(@floatFromInt(rhs_fixed))),

			else => self.raise(error.TypeMismatch, "fixed types are not addable with {typename}", .{
				.typename = addable.typename(),
			}),
		};
	}

	///
	/// Attempts to pop the top-most value in `self` and divide it with `rhs_fixed` to it, pushing the result.
	///
	/// A [RuntimeError] is returned if `self` is out of memory or `rhs_fixed` is not divisible with the top-most local.
	///
	/// `self` is returned for function chaining.
	///
	pub fn fixed_divide(self: *RuntimeEnv, rhs_fixed: Fixed) RuntimeError!*RuntimeEnv {
		if (rhs_fixed == 0) {
			return self.raise(error.TypeMismatch, "cannot divide by zero", .{});
		}

		const addable = try self.expect_object(self.pop());

		defer self.release(addable);

		return switch (addable.internal().payload) {
			.fixed => |lhs_fixed| self.new_float(@as(Float, @floatFromInt(lhs_fixed)) / @as(Float, @floatFromInt(rhs_fixed))),
			.float => |lhs_float| self.new_float(lhs_float / @as(Float, @floatFromInt(rhs_fixed))),
			.vector2 => |lhs_vector2| self.new_vector2(lhs_vector2.scalar_divided(@floatFromInt(rhs_fixed))),
			.vector3 => |lhs_vector3| self.new_vector3(lhs_vector3.scalar_divided(@floatFromInt(rhs_fixed))),

			else => self.raise(error.TypeMismatch, "fixed types are not divisible with {typename}", .{
				.typename = addable.typename(),
			}),
		};
	}

	///
	/// Attempts to pop the top-most value in `self` and multiply it with `rhs_fixed` to it, pushing the result.
	///
	/// A [RuntimeError] is returned if `self` is out of memory or `rhs_fixed` is not multiplicable with the top-most
	/// local.
	///
	/// `self` is returned for function chaining.
	///
	pub fn fixed_multiply(self: *RuntimeEnv, rhs_fixed: Fixed) RuntimeError!*RuntimeEnv {
		const addable = try self.expect_object(self.pop());

		defer self.release(addable);

		return switch (addable.internal().payload) {
			.fixed => |lhs_fixed| if (coral.math.checked_mul(lhs_fixed, rhs_fixed)) |result|
					self.new_fixed(result)
				else
					self.new_float(@as(Float, @floatFromInt(lhs_fixed)) * @as(Float, @floatFromInt(rhs_fixed))),

			.float => |lhs_float| self.new_float(lhs_float + @as(Float, @floatFromInt(rhs_fixed))),
			.vector2 => |lhs_vector2| self.new_vector2(lhs_vector2.scalar_multiplied(@floatFromInt(rhs_fixed))),
			.vector3 => |lhs_vector3| self.new_vector3(lhs_vector3.scalar_multiplied(@floatFromInt(rhs_fixed))),

			else => self.raise(error.TypeMismatch, "fixed types are not multiplicable with {typename}", .{
				.typename = addable.typename(),
			}),
		};
	}

	///
	/// Attempts to pop the top-most value in `self` and subtract it with `rhs_fixed`, pushing the result.
	///
	/// A [RuntimeError] is returned if `self` is out of memory or `rhs_fixed` is not subtractable with the top-most
	/// local.
	///
	/// `self` is returned for function chaining.
	///
	pub fn fixed_subtract(self: *RuntimeEnv, rhs_fixed: Fixed) RuntimeError!*RuntimeEnv {
		const addable = try self.expect_object(self.pop());

		defer self.release(addable);

		return switch (addable.internal().payload) {
			.fixed => |lhs_fixed| if (coral.math.checked_sub(lhs_fixed, rhs_fixed)) |result|
					self.new_fixed(result)
				else
					self.new_float(@as(Float, @floatFromInt(lhs_fixed)) - @as(Float, @floatFromInt(rhs_fixed))),

			.float => |lhs_float| self.new_float(lhs_float + @as(Float, @floatFromInt(rhs_fixed))),
			.vector2 => |lhs_vector2| self.new_vector2(lhs_vector2.scalar_subtracted(@floatFromInt(rhs_fixed))),
			.vector3 => |lhs_vector3| self.new_vector3(lhs_vector3.scalar_subtracted(@floatFromInt(rhs_fixed))),

			else => self.raise(error.TypeMismatch, "fixed types are not multiplicable with {typename}", .{
				.typename = addable.typename(),
			}),
		};
	}

	///
	/// Attempts to pop the top-most value in `self` and add it with `rhs_float`, pushing the result.
	///
	/// A [RuntimeError] is returned if `self` is out of memory or `rhs_float` is not addable with the top-most local.
	///
	/// `self` is returned for function chaining.
	///
	pub fn float_add(self: *RuntimeEnv, rhs_float: Float) RuntimeError!*RuntimeEnv {
		const addable = try self.expect_object(self.pop());

		defer self.release(addable);

		return switch (addable.internal().payload) {
			.fixed => |lhs_fixed| self.new_float(@as(Float, @floatFromInt(lhs_fixed)) + rhs_float),
			.float => |lhs_float| self.new_float(lhs_float + lhs_float),
			.vector2 => |lhs_vector2| self.new_vector2(lhs_vector2.scalar_added(@floatCast(rhs_float))),
			.vector3 => |lhs_vector3| self.new_vector3(lhs_vector3.scalar_added(@floatCast(rhs_float))),

			else => self.raise(error.TypeMismatch, "fixed types are not addable with {typename}", .{
				.typename = addable.typename(),
			}),
		};
	}

	///
	/// Attempts to pop the top-most value in `self` and divide it with `rhs_float` to it, pushing the result.
	///
	/// A [RuntimeError] is returned if `self` is out of memory or `rhs_float` is not divisible with the top-most local.
	///
	/// `self` is returned for function chaining.
	///
	pub fn float_divide(self: *RuntimeEnv, rhs_float: Float) RuntimeError!*RuntimeEnv {
		if (rhs_float == 0) {
			return self.raise(error.TypeMismatch, "cannot divide by zero", .{});
		}

		const addable = try self.expect_object(self.pop());

		defer self.release(addable);

		return switch (addable.internal().payload) {
			.fixed => |lhs_fixed| self.new_float(@as(Float, @floatFromInt(lhs_fixed)) / rhs_float),
			.float => |lhs_float| self.new_float(lhs_float / lhs_float),
			.vector2 => |lhs_vector2| self.new_vector2(lhs_vector2.scalar_divided(@floatCast(rhs_float))),
			.vector3 => |lhs_vector3| self.new_vector3(lhs_vector3.scalar_divided(@floatCast(rhs_float))),

			else => self.raise(error.TypeMismatch, "fixed types are not divisible with {typename}", .{
				.typename = addable.typename(),
			}),
		};
	}

	///
	/// Attempts to pop the top-most value in `self` and multiply it with `rhs_float`, pushing the result.
	///
	/// A [RuntimeError] is returned if `self` is out of memory or `rhs_float` is not multiplicable with the top-most
	/// local.
	///
	/// `self` is returned for function chaining.
	///
	pub fn float_multiply(self: *RuntimeEnv, rhs_float: Float) RuntimeError!*RuntimeEnv {
		const addable = try self.expect_object(self.pop());

		defer self.release(addable);

		return switch (addable.internal().payload) {
			.fixed => |lhs_fixed| self.new_float(@as(Float, @floatFromInt(lhs_fixed)) * rhs_float),
			.float => |lhs_float| self.new_float(lhs_float * lhs_float),
			.vector2 => |lhs_vector2| self.new_vector2(lhs_vector2.scalar_multiplied(@floatCast(rhs_float))),
			.vector3 => |lhs_vector3| self.new_vector3(lhs_vector3.scalar_multiplied(@floatCast(rhs_float))),

			else => self.raise(error.TypeMismatch, "fixed types are not multiplicable with {typename}", .{
				.typename = addable.typename(),
			}),
		};
	}

	///
	/// Attempts to pop the top-most value in `self` and subtract it with `rhs_float`, pushing the result.
	///
	/// A [RuntimeError] is returned if `self` is out of memory or `rhs_float` is not subtractable with the top-most
	/// local.
	///
	/// `self` is returned for function chaining.
	///
	pub fn float_subtract(self: *RuntimeEnv, rhs_float: Float) RuntimeError!*RuntimeEnv {
		const addable = try self.expect_object(self.pop());

		defer self.release(addable);

		return switch (addable.internal().payload) {
			.fixed => |lhs_fixed| self.new_float(@as(Float, @floatFromInt(lhs_fixed)) - rhs_float),
			.float => |lhs_float| self.new_float(lhs_float - lhs_float),
			.vector2 => |lhs_vector2| self.new_vector2(lhs_vector2.scalar_subtracted(@floatCast(rhs_float))),
			.vector3 => |lhs_vector3| self.new_vector3(lhs_vector3.scalar_subtracted(@floatCast(rhs_float))),

			else => self.raise(error.TypeMismatch, "fixed types are not multiplicable with {typename}", .{
				.typename = addable.typename(),
			}),
		};
	}

	///
	/// Attempts to pop the top-most value and index into it with `index`, pushing the retrieved value.
	///
	/// A [RuntimeError] is returned if `self` is out of memory, the top-most local is nil or not indexable, or the
	/// object-specific indexing operation raised an error.
	///
	/// `self` is returned for function chaining.
	///
	pub fn index_get(self: *RuntimeEnv, index: *RuntimeObj) RuntimeError!*RuntimeEnv {
		const vector_max = 3;
		const indexable = try self.expect_object(self.pop());

		defer self.release(indexable);

		switch (indexable.internal().payload) {
			.vector2 => |vector2| {
				const swizzle_symbol = try self.expect_symbol(index);
				var swizzle_buffer = [_]f32{0} ** vector_max;
				var swizzle_count = @as(usize, 0);

				while (true) : (swizzle_count += 1) {
					if (swizzle_count > swizzle_buffer.len) {
						try self.locals.push_one(null);

						return self;
					}

					swizzle_buffer[swizzle_count] = switch (swizzle_symbol[swizzle_count]) {
						'x' => vector2.x,
						'y' => vector2.y,

						0 => return switch (swizzle_count) {
							1 => self.new_float(swizzle_buffer[0]),

							2 => self.new_vector2(.{
								.x = swizzle_buffer[0],
								.y = swizzle_buffer[1],
							}),

							3 => self.new_vector3(Vector3.from_scalars(swizzle_buffer)),
							else => unreachable,
						},

						else => return self.raise(error.BadOperation, "no such vector2 index or swizzle: `{index}`", .{
							.index = coral.io.slice_sentineled(@as(coral.io.Byte, 0), swizzle_symbol),
						}),
					};
				}
			},

			.vector3 => |vector3| {
				const swizzle_symbol = try self.expect_symbol(index);
				var swizzle_buffer = [_]f32{0} ** vector_max;
				var swizzle_count = @as(usize, 0);

				while (true) : (swizzle_count += 1) {
					if (swizzle_count > swizzle_buffer.len) {
						try self.locals.push_one(null);

						return self;
					}

					swizzle_buffer[swizzle_count] = switch (swizzle_symbol[swizzle_count]) {
						'x' => vector3.x,
						'y' => vector3.y,
						'z' => vector3.z,

						0 => return switch (swizzle_count) {
							1 => self.new_float(swizzle_buffer[0]),

							2 => self.new_vector2(.{
								.x = swizzle_buffer[0],
								.y = swizzle_buffer[1],
							}),

							3 => self.new_vector3(Vector3.from_scalars(swizzle_buffer)),
							else => unreachable,
						},

						else => return self.raise(error.BadOperation, "no such vector3 index or swizzle: `{index}`", .{
							.index = coral.io.slice_sentineled(@as(coral.io.Byte, 0), swizzle_symbol),
						}),
					};
				}
			},

			.dynamic => |dynamic| {
				return self.push(get_value: {
					try self.locals.push_one(index);

					defer coral.debug.assert(self.locals.drop(1));

					break: get_value try dynamic.typeinfo().get(.{
						.userdata = dynamic.userdata(),
						.env = self,
					});
				});
			},

			else => return self.raise(error.TypeMismatch, "{typename} is not get-indexable", .{
				.typename = indexable.typename(),
			}),
		}
	}

	///
	/// Attempts to pop the top-most value and index into it with `index`, inserting `value` into it.
	///
	/// A [RuntimeError] is returned if `self` is out of memory, the top-most local is nil or not indexable, or the
	/// object-specific indexing operation raised an error.
	///
	pub fn index_set(self: *RuntimeEnv, index: *RuntimeObj, value: ?*RuntimeObj) RuntimeError!void {
		const indexable = try self.expect_object(self.pop());

		defer self.release(indexable);

		try switch (indexable.internal().payload) {
			.dynamic => |dynamic| set_value: {
				try self.locals.push_one(value);

				defer coral.debug.assert(self.locals.pop() != null);

				try self.locals.push_one(index);

				defer coral.debug.assert(self.locals.pop() != null);

				break: set_value dynamic.typeinfo().set(.{
					.userdata = dynamic.userdata(),
					.env = self,
				});
			},

			else => self.raise(error.TypeMismatch, "{typename} is not set-indexable", .{
				.typename = indexable.typename(),
			}),
		};
	}

	///
	/// Attempts to import a script from the import [file.Access] (specified during [RuntimeEnv.init]) from `file_path`.
	///
	/// If the loaded script is a plain text file, it will be read, parsed, and compiled into an optimized format before
	/// being executed.
	///
	/// If the loaded script is already compiled, it is immediately executed.
	///
	/// After completing execution, the return value is pushed to the top of `self`.
	///
	/// A [RuntimeError] is returned if `self` is out of memory, no file at `file_path` could be found by the runtime,
	/// the script is a source file and contains invalid syntax, or the imported script raised an error during
	/// execution.
	///
	/// Any syntax errors are reported using [print_error] prior to raising a runtime error.
	///
	/// `self` is returned for function chaining.
	///
	pub fn import(self: *RuntimeEnv, file_path: file.Path) RuntimeError!*RuntimeEnv {
		{
			var callable = new_chunk: {
				const file_name = file_path.get_string();

				const file_data =
					(try file.allocate_and_load(self.allocator, self.options.import_access, file_path)) orelse {
						return self.raise(error.BadOperation, "failed to open or read `{name}`", .{
							.name = file_name,
						});
					};

				defer self.allocator.deallocate(file_data);

				var root = try tree.Root.init(self.allocator);

				defer root.deinit();

				{
					var stream = tokens.Stream{.source = file_data};

					root.parse(&stream) catch |parse_error| {
						for (root.error_messages.values) |error_message| {
							self.print_error(error_message);
						}

						return self.raise(parse_error, "failed to parse `{name}`", .{.name = file_name});
					};
				}

				var chunk = try Chunk.init(self, file_name, &root.environment);

				errdefer chunk.deinit(self);

				break: new_chunk (try self.new_dynamic(coral.io.bytes_of(&chunk), Chunk.typeinfo)).pop().?;
			};

			errdefer self.release(callable);

			try self.locals.push_one(callable);
		}

		return self.call(0);
	}

	///
	/// Attempts to initialize a new virtual machine instance with `allocator` as the allocation strategy, `frames_max`
	/// as the maximum call depth, and `options` as the optional values to pass to the VM state.
	///
	/// This function will fail if `allocator` cannot allocate the memory required to create the runtime environment.
	///
	pub fn init(allocator: coral.io.Allocator, frames_max: u32, options: Options) coral.io.AllocationError!RuntimeEnv {
		var frames = FrameList.init(allocator);

		errdefer frames.deinit();

		try frames.grow(frames_max);

		var locals = LocalList.init(allocator);

		errdefer locals.deinit();

		try locals.grow(local_max * frames_max);

		return .{
			.interned_symbols = SymbolSet.init(allocator, .{}),
			.locals = locals,
			.frames = frames,
			.allocator = allocator,
			.options = options,
		};
	}

	const Local = u8;

	const local_max = coral.math.max_int(@typeInfo(Local).Int);

	///
	/// Attempts to push the local located at `index` to the top of `self`.
	///
	/// Locals are indexed according to the order they are declared, with `0` referring to the first declared arg, let,
	/// or var in the current frame.
	///
	/// A [RuntimeError] is returned if `self` is out of memory, `index` does not refer to a valid local, or the
	/// virtual-machine is not in a managed call frame.
	///
	/// `self` is returned for function chaining.
	///
	pub fn local_get(self: *RuntimeEnv, index: Local) RuntimeError!*RuntimeEnv {
		const frame = self.frames.peek() orelse {
			return self.raise(error.IllegalState, "cannot get locals outside of a call frame", .{});
		};

		const locals = frame.locals_of(self);

		if (index >= locals.len) {
			return self.raise(error.IllegalState, "invalid local get", .{});
		}

		return self.push(locals[index]);
	}

	///
	/// Attempts to set the local located at `index` to `value`.
	///
	/// Locals are indexed according to the order they are declared, with `0` referring to the first declared arg, let,
	/// or var in the current frame.
	///
	/// A [RuntimeError] if`index` does not refer to a valid local, or the virtual-machine is not in a managed call
	/// frame.
	///
	/// `self` is returned for function chaining.
	///
	pub fn local_set(self: *RuntimeEnv, index: Local, value: ?*RuntimeObj) RuntimeError!void {
		const frame = self.frames.peek() orelse {
			return self.raise(error.IllegalState, "cannot get locals outside of a call frame", .{});
		};

		const locals = frame.locals_of(self);

		if (index >= locals.len) {
			return self.raise(error.IllegalState, "invalid local get", .{});
		}

		const local = &locals[index];

		if (local.*) |previous_local| {
			self.release(previous_local);
		}

		local.* = if (value) |object| object.internal().acquire() else null;
	}

	///
	/// Attempts to push a copy of the top-most local in `self`.
	///
	/// A [RuntimeError] is returned if `self` is out of memory, the virtual-machine is not in a managed call frame, or
	/// there are no locals in the frame.
	///
	/// `self` is returned for function chaining.
	///
	pub fn local_top(self: *RuntimeEnv) RuntimeError!*RuntimeEnv {
		const frame = self.frames.peek() orelse {
			return self.raise(error.IllegalState, "cannot get locals outside of a call frame", .{});
		};

		const locals = frame.locals_of(self);

		if (locals.len != 0) {
			return self.raise(error.IllegalState, "invalid local top", .{});
		}

		return self.push(locals[locals.len - 1]);
	}

	///
	/// Pops the top-most local from `self`, negates it, and pushes the result.
	///
	/// A [RuntimeError] is returned if `self` is out of memory or the top-most local is nil or not negatable.
	///
	/// `self` is returned for function chaining.
	///
	pub fn neg(self: *RuntimeEnv) RuntimeError!*RuntimeEnv {
		const negatable = try self.expect_object(self.pop());

		defer self.release(negatable);

		return switch (negatable.internal().payload) {
			.fixed => |fixed| self.new_fixed(-fixed),
			.float => |float| self.new_float(-float),

			else => self.raise(error.TypeMismatch, "{typename} is not scalar negatable", .{
				.typename = negatable.typename(),
			}),
		};
	}

	///
	/// Attempts to create a new true or false object according to `value`, pushing it to the top of `self`.
	///
	/// A [RuntimeError] is returned if `self` is out of memory.
	///
	/// `self` is returned for function chaining.
	///
	pub fn new_boolean(self: *RuntimeEnv, value: bool) RuntimeError!*RuntimeEnv {
		const boolean = try RuntimeObj.allocate(self.allocator, .{
			.ref_count = 1,
			.payload = if (value) .true else .false,
		});

		errdefer self.release(boolean);

		try self.locals.push_one(boolean);

		return self;
	}

	///
	/// Attempts to create a new boxed object holding a reference to `value`, pushing it to the top of `self`.
	///
	/// A [RuntimeError] is returned if `self` is out of memory.
	///
	/// `self` is returned for function chaining.
	///
	pub fn new_boxed(self: *RuntimeEnv, value: ?*RuntimeObj) RuntimeError!*RuntimeEnv {
		const boxed = try RuntimeObj.allocate(self.allocator, .{
			.ref_count = 1,
			.payload = .{.boxed = if (value) |object| object.internal().acquire() else null},
		});

		errdefer self.release(boxed);

		try self.locals.push_one(boxed);

		return self;
	}

	///
	/// Attempts to create a new dynamic object from `userdata` and `typeinfo`, pushing it to the top of `self`.
	///
	/// *Note* that the number of bytes in `userdata` must match the size described in `typeinfo` exactly.
	///
	/// A [RuntimeError] is returned if `self` is out of memory.
	///
	/// `self` is returned for function chaining.
	///
	pub fn new_dynamic(self: *RuntimeEnv, userdata: []const coral.io.Byte, typeinfo: *const Typeinfo) RuntimeError!*RuntimeEnv {
		coral.debug.assert(userdata.len == typeinfo.size);

		const dynamic = new: {
			const dynamic = try self.allocator.reallocate(null, @sizeOf(usize) + typeinfo.size);

			errdefer self.allocator.deallocate(dynamic);

			coral.io.copy(dynamic, coral.io.bytes_of(&typeinfo));
			coral.io.copy(dynamic[@sizeOf(usize) ..], userdata[0 .. typeinfo.size]);

			break: new try RuntimeObj.allocate(self.allocator, .{
				.ref_count = 1,
				.payload = .{.dynamic = .{.ptr = dynamic.ptr}},
			});
		};

		errdefer self.release(dynamic);

		try self.locals.push_one(dynamic);

		return self;
	}

	///
	/// Attempts to create a new fixed-size number object from `value`, pushing it to the top of `self`.
	///
	/// A [RuntimeError] is returned if `self` is out of memory.
	///
	/// `self` is returned for function chaining.
	///
	pub fn new_fixed(self: *RuntimeEnv, value: Fixed) RuntimeError!*RuntimeEnv {
		const syscall = try RuntimeObj.allocate(self.allocator, .{
			.ref_count = 1,
			.payload = .{.fixed = value},
		});

		errdefer self.release(syscall);

		try self.locals.push_one(syscall);

		return self;
	}

	///
	/// Attempts to create a new floating point number object from `value`, pushing it to the top of `self`.
	///
	/// A [RuntimeError] is returned if `self` is out of memory.
	///
	/// `self` is returned for function chaining.
	///
	pub fn new_float(self: *RuntimeEnv, value: Float) RuntimeError!*RuntimeEnv {
		const syscall = try RuntimeObj.allocate(self.allocator, .{
			.ref_count = 1,
			.payload = .{.float = value},
		});

		errdefer self.release(syscall);

		try self.locals.push_one(syscall);

		return self;
	}

	///
	/// Attempts to create a new byte string object from `value`, pushing it to the top of `self`.
	///
	/// A [RuntimeError] is returned if `self` is out of memory.
	///
	/// `self` is returned for function chaining.
	///
	pub fn new_string(self: *RuntimeEnv, value: []const coral.io.Byte) RuntimeError!*RuntimeEnv {
		if (value.len > coral.math.max_int(@typeInfo(Fixed).Int)) {
			return error.OutOfMemory;
		}

		const string = new: {
			const string = try coral.io.allocate_copy(coral.io.Byte, self.allocator, value);

			errdefer self.allocator.deallocate(string);

			break: new try RuntimeObj.allocate(self.allocator, .{
				.payload = .{
					.string = .{
						.ptr = string.ptr,
						.len = @intCast(string.len),
					},
				},

				.ref_count = 1,
			});
		};

		errdefer self.release(string);

		try self.locals.push_one(string);

		return self;
	}

	///
	/// Attempts to create a new system-native function object from `value`, pushing it to the top of `self`.
	///
	/// A [RuntimeError] is returned if `self` is out of memory.
	///
	/// `self` is returned for function chaining.
	///
	pub fn new_syscall(self: *RuntimeEnv, value: *const Syscall) RuntimeError!*RuntimeEnv {
		const syscall = try RuntimeObj.allocate(self.allocator, .{
			.ref_count = 1,
			.payload = .{.syscall = value},
		});

		errdefer self.release(syscall);

		try self.locals.push_one(syscall);

		return self;
	}

	///
	/// Attempts to create a new empty table object, pushing it to the top of `self`.
	///
	/// A [RuntimeError] is returned if `self` is out of memory.
	///
	/// `self` is returned for function chaining.
	///
	pub fn new_table(self: *RuntimeEnv) RuntimeError!*RuntimeEnv {
		var table = Table.init(self);

		errdefer table.deinit(self);

		return self.new_dynamic(coral.io.bytes_of(&table), Table.typeinfo);
	}

	///
	/// Attempts to create a new 2-component vector object from `value`, pushing it to the top of `self`.
	///
	/// A [RuntimeError] is returned if `self` is out of memory.
	///
	/// `self` is returned for function chaining.
	///
	pub fn new_vector2(self: *RuntimeEnv, value: Vector2) RuntimeError!*RuntimeEnv {
		const vector2 = try RuntimeObj.allocate(self.allocator, .{
			.ref_count = 1,
			.payload = .{.vector2 = value},
		});

		errdefer self.release(vector2);

		try self.locals.push_one(vector2);

		return self;
	}

	///
	/// Attempts to create a new 3-component vector object from `value`, pushing it to the top of `self`.
	///
	/// A [RuntimeError] is returned if `self` is out of memory.
	///
	/// `self` is returned for function chaining.
	///
	pub fn new_vector3(self: *RuntimeEnv, value: Vector3) RuntimeError!*RuntimeEnv {
		const vector3 = try RuntimeObj.allocate(self.allocator, .{
			.ref_count = 1,
			.payload = .{.vector3 = value},
		});

		errdefer self.release(vector3);

		try self.locals.push_one(vector3);

		return self;
	}

	///
	/// Attempts to create a new symbol object from `value`, pushing it to the top of `self`.
	///
	/// A [RuntimeError] is returned if `self` is out of memory.
	///
	/// `self` is returned for function chaining.
	///
	pub fn new_symbol(self: *RuntimeEnv, value: []const coral.io.Byte) RuntimeError!*RuntimeEnv {
		const symbol = try RuntimeObj.allocate(self.allocator, .{
			.payload = .{
				.symbol = self.interned_symbols.lookup(value) orelse create: {
					const symbol_string = try coral.io.allocate_string(self.allocator, value);

					errdefer self.allocator.deallocate(symbol_string);

					coral.debug.assert(try self.interned_symbols.insert(symbol_string[0 .. value.len], symbol_string));

					break: create symbol_string;
				},
			},

			.ref_count = 1,
		});

		errdefer self.release(symbol);

		try self.locals.push_one(symbol);

		return self;
	}

	///
	/// Pops the top-most value of `self`, returning the value.
	///
	/// *Note* any returned non-null pointer must be released via [RuntimeEnv.release] or else the resources belonging
	/// to the objects will never be freed by the runtime.
	///
	pub fn pop(self: *RuntimeEnv) ?*RuntimeObj {
		const popped = self.locals.pop();

		coral.debug.assert(popped != null);

		return popped.?;
	}

	///
	/// Prints `buffer` to the printing operation specified in the options of `self`.
	///
	pub fn print(self: *RuntimeEnv, buffer: []const coral.io.Byte) void {
		if (self.options.print) |op| {
			op(buffer);
		}
	}

	///
	/// Prints `buffer` to the error printing operation specified in the options of `self.
	///
	pub fn print_error(self: *RuntimeEnv, buffer: []const coral.io.Byte) void {
		if (self.options.print_error) |op| {
			op(buffer);
		}
	}

	///
	/// Attempts to push `value` to the top of `self`.
	///
	/// A [RuntimeError] is returned if `self` is out of memory.
	///
	/// `self` is returned for function chaining.
	///
	pub fn push(self: *RuntimeEnv, value: ?*RuntimeObj) RuntimeError!*RuntimeEnv {
		if (value) |object| {
			const acquired = object.internal().acquire();

			errdefer self.release(acquired);

			try self.locals.push_one(acquired);
		} else {
			try self.locals.push_one(null);
		}

		return self;
	}

	///
	/// Creates stack-unwinding runtime error from `error_value` with `format` and `args` as the formatted error message
	/// presented with the type of error raised and a full stack-trace.
	///
	/// The `error_value` passed is returned so that the error may be raised and returned in a single line.
	///
	pub fn raise(self: *RuntimeEnv, error_value: RuntimeError, comptime format: []const coral.io.Byte, args: anytype) RuntimeError {
		const formatted_message = try coral.utf8.alloc_formatted(self.allocator, format, args);

		defer self.allocator.deallocate(formatted_message);

		self.print_error(formatted_message);

		if (!self.frames.is_empty()) {
			self.print_error("stack trace:");

			var remaining_frames = self.frames.values.len;

			while (remaining_frames != 0) {
				remaining_frames -= 1;

				const callable = self.frames.values[remaining_frames].callable;
				const name = (try (try self.push(callable)).to_string()).pop().?;

				defer self.release(name);

				if (callable.is_dynamic(Chunk.typeinfo)) |chunk_userdata| {
					const chunk = @as(*Chunk, @ptrCast(@alignCast(chunk_userdata)));
					const line = chunk.lines.values[chunk.cursor];

					const chunk_name = try coral.utf8.alloc_formatted(self.allocator, "{name}@{line_number}", .{
						.name = get_name: {
							const string = name.is_string();

							coral.debug.assert(string != null);

							break: get_name string.?;
						},

						.line_number = line.number,
					});

					defer self.allocator.deallocate(chunk_name);

					self.print_error(chunk_name);
				} else {
					self.print_error(get_name: {
						const string = name.is_string();

						coral.debug.assert(string != null);

						break: get_name string.?;
					});
				}
			}
		}

		return error_value;
	}

	///
	/// Releases `value` from `self`, decrementing the reference count (if applicable), and releasing all resources
	/// belonging to the object once the reference count reaches zero.
	///
	pub fn release(self: *RuntimeEnv, value: *RuntimeObj) void {
		coral.debug.assert(value.internal().ref_count != 0);

		value.internal().ref_count -= 1;

		if (value.internal().ref_count == 0) {
			switch (value.internal().payload) {
				.false, .true, .float, .fixed, .symbol, .vector2, .vector3, .syscall => {},

				.boxed => |*boxed| {
					if (boxed.*) |boxed_value| {
						self.release(boxed_value);
					}
				},

				.string => |string| {
					coral.debug.assert(string.len >= 0);
					self.allocator.deallocate(string.ptr[0 .. @intCast(string.len)]);
				},

				.dynamic => |dynamic| {
					if (dynamic.typeinfo().destruct) |destruct| {
						destruct(.{
							.userdata = dynamic.userdata(),
							.env = self,
						});
					}

					self.allocator.deallocate(dynamic.unpack());
				},
			}

			self.allocator.deallocate(value.internal());
		}
	}

	///
	/// Attempts to pop the top-most value of `self`, convert it to a string object representation, and push the result.
	///
	/// A [RuntimeError] is returned if `self` is out of memory or the to_string implementation of the top-most value of
	/// `self` raised an error.
	///
	/// `self` is returned for function chaining.
	///
	pub fn to_string(self: *RuntimeEnv) RuntimeError!*RuntimeEnv {
		const decimal_format = coral.utf8.DecimalFormat.default;
		const value = try self.expect_object(self.pop());

		defer self.release(value);

		return switch (value.internal().payload) {
			.false => self.new_string("false"),
			.true => self.new_string("true"),

			.fixed => |fixed| convert: {
				var string = [_:0]coral.io.Byte{0} ** 32;
				var buffer = coral.io.FixedBuffer{.bytes = &string};
				const length = decimal_format.print(buffer.as_writer(), fixed);

				coral.debug.assert(length != null);

				break: convert self.new_string(string[0 .. length.?]);
			},

			.float => |float| convert: {
				var string = [_:0]coral.io.Byte{0} ** 32;
				var buffer = coral.io.FixedBuffer{.bytes = &string};
				const length = decimal_format.print(buffer.as_writer(), float);

				coral.debug.assert(length != null);

				break: convert self.new_string(string[0 .. length.?]);
			},

			.boxed => |boxed| (try self.push(try self.expect_object(boxed))).to_string(),
			.symbol => |symbol| self.new_string(coral.io.slice_sentineled(@as(coral.io.Byte, 0), symbol)),

			.string => acquire: {
				try self.locals.push_one(value.internal().acquire());

				break: acquire self;
			},

			.vector2 => |vector2| convert: {
				var string = [_:0]coral.io.Byte{0} ** 64;
				var buffer = coral.io.FixedBuffer{.bytes = &string};
				const length = coral.utf8.print_formatted(buffer.as_writer(), "@vec2({x}, {y})", vector2).?;

				break: convert self.new_string(string[0 .. length]);
			},

			.vector3 => |vector3| convert: {
				var string = [_:0]coral.io.Byte{0} ** 96;
				var buffer = coral.io.FixedBuffer{.bytes = &string};
				const length = coral.utf8.print_formatted(buffer.as_writer(), "@vec3({x}, {y}, {z})", vector3).?;

				break: convert self.new_string(string[0 .. length]);
			},

			.syscall => self.new_string("syscall"),

			.dynamic => |dynamic| dynamic_to_string: {
				const string = try dynamic.typeinfo().to_string(.{
					.userdata = dynamic.userdata(),
					.env = self,
				});

				errdefer self.release(string);

				try self.locals.push_one(string);

				break: dynamic_to_string self;
			},
		};
	}

	///
	/// Attempts to pop the top-most value in `self` and add it with `rhs_vector2`, pushing the result.
	///
	/// A [RuntimeError] is returned if `self` is out of memory or `rhs_vector2` is not addable with the top-most local.
	///
	/// `self` is returned for function chaining.
	///
	pub fn vector2_add(self: *RuntimeEnv, rhs_vector2: Vector2) RuntimeError!*RuntimeEnv {
		const addable = try self.expect_object(self.pop());

		defer self.release(addable);

		return if (addable.is_vector2()) |lhs_vector2|
				self.new_vector2(lhs_vector2.vector_added(rhs_vector2))
			else
				self.raise(error.TypeMismatch, "vector2 types are not addable with {typename}", .{
					.typename = addable.typename(),
				});
	}

	///
	/// Attempts to pop the top-most value in `self` and divide it with `rhs_vector2`, pushing the result.
	///
	/// A [RuntimeError] is returned if `self` is out of memory or `rhs_vector2` is not divisible with the top-most
	/// local.
	///
	/// `self` is returned for function chaining.
	///
	pub fn vector2_divide(self: *RuntimeEnv, rhs_vector2: Vector2) RuntimeError!*RuntimeEnv {
		if (rhs_vector2.x == 0 or rhs_vector2.y == 0) {
			return self.raise(error.TypeMismatch, "cannot divide by zero", .{});
		}

		const addable = try self.expect_object(self.pop());

		defer self.release(addable);

		return if (addable.is_vector2()) |lhs_vector2|
				self.new_vector2(lhs_vector2.vector_divided(rhs_vector2))
			else
				self.raise(error.TypeMismatch, "vector2 types are not divisible with {typename}", .{
					.typename = addable.typename(),
				});
	}

	///
	/// Attempts to pop the top-most value in `self` and multiply it with `rhs_vector2`, pushing the result.
	///
	/// A [RuntimeError] is returned if `self` is out of memory or `rhs_vector2` is not multiplicable with the top-most
	/// local.
	///
	/// `self` is returned for function chaining.
	///
	pub fn vector2_multiply(self: *RuntimeEnv, rhs_vector2: Vector2) RuntimeError!*RuntimeEnv {
		const addable = try self.expect_object(self.pop());

		defer self.release(addable);

		return if (addable.is_vector2()) |lhs_vector2|
				self.new_vector2(lhs_vector2.vector_multiplied(rhs_vector2))
			else
				self.raise(error.TypeMismatch, "vector2 types are not multiplicable with {typename}", .{
					.typename = addable.typename(),
				});
	}

	///
	/// Attempts to pop the top-most value in `self` and subtract it with `rhs_vector2`, pushing the result.
	///
	/// A [RuntimeError] is returned if `self` is out of memory or `rhs_vector2` is not subtractable with the top-most
	/// local.
	///
	/// `self` is returned for function chaining.
	///
	pub fn vector2_subtract(self: *RuntimeEnv, rhs_vector2: Vector2) RuntimeError!*RuntimeEnv {
		const addable = try self.expect_object(self.pop());

		defer self.release(addable);

		return if (addable.is_vector2()) |lhs_vector2|
				self.new_vector2(lhs_vector2.vector_subtracted(rhs_vector2))
			else
				self.raise(error.TypeMismatch, "vector2 types are not multiplicable with {typename}", .{
					.typename = addable.typename(),
				});
	}

	///
	/// Attempts to pop the top-most value in `self` and add it with `rhs_vector3`, pushing the result.
	///
	/// A [RuntimeError] is returned if `self` is out of memory or `rhs_vector3` is not addable with the top-most local.
	///
	/// `self` is returned for function chaining.
	///
	pub fn vector3_add(self: *RuntimeEnv, rhs_vector3: Vector3) RuntimeError!*RuntimeEnv {
		const addable = try self.expect_object(self.pop());

		defer self.release(addable);

		return if (addable.is_vector3()) |lhs_vector3|
				self.new_vector3(lhs_vector3.vector_added(rhs_vector3))
			else
				self.raise(error.TypeMismatch, "vector3 types are not addable with {typename}", .{
					.typename = addable.typename(),
				});
	}

	///
	/// Attempts to pop the top-most value in `self` and divide it with `rhs_vector3`, pushing the result.
	///
	/// A [RuntimeError] is returned if `self` is out of memory or `rhs_vector3` is not divisible with the top-most
	/// local.
	///
	/// `self` is returned for function chaining.
	///
	pub fn vector3_divide(self: *RuntimeEnv, rhs_vector3: Vector3) RuntimeError!*RuntimeEnv {
		if (rhs_vector3.x == 0 or rhs_vector3.y == 0 or rhs_vector3.z == 0) {
			return self.raise(error.TypeMismatch, "cannot divide by zero", .{});
		}

		const addable = try self.expect_object(self.pop());

		defer self.release(addable);

		return if (addable.is_vector3()) |lhs_vector3|
				self.new_vector3(lhs_vector3.vector_divided(rhs_vector3))
			else
				self.raise(error.TypeMismatch, "vector3 types are not divisible with {typename}", .{
					.typename = addable.typename(),
				});
	}

	///
	/// Attempts to pop the top-most value in `self` and multiply it with `rhs_vector3`, pushing the result.
	///
	/// A [RuntimeError] is returned if `self` is out of memory or `rhs_vector3` is not multiplicable with the top-most
	/// local.
	///
	/// `self` is returned for function chaining.
	///
	pub fn vector3_multiply(self: *RuntimeEnv, rhs_vector3: Vector3) RuntimeError!*RuntimeEnv {
		const addable = try self.expect_object(self.pop());

		defer self.release(addable);

		return if (addable.is_vector3()) |lhs_vector3|
				self.new_vector3(lhs_vector3.vector_multiplied(rhs_vector3))
			else
				self.raise(error.TypeMismatch, "vector3 types are not multiplicable with {typename}", .{
					.typename = addable.typename(),
				});
	}

	///
	/// Attempts to pop the top-most value in `self` and subtract it with `rhs_vector3`, pushing the result.
	///
	/// A [RuntimeError] is returned if `self` is out of memory or `rhs_vector3` is not subtractable with the top-most
	/// local.
	///
	/// `self` is returned for function chaining.
	///
	pub fn vector3_subtract(self: *RuntimeEnv, rhs_vector3: Vector3) RuntimeError!*RuntimeEnv {
		const addable = try self.expect_object(self.pop());

		defer self.release(addable);

		return if (addable.is_vector3()) |lhs_vector3|
				self.new_vector3(lhs_vector3.vector_subtracted(rhs_vector3))
			else
				self.raise(error.TypeMismatch, "vector3 types are not multiplicable with {typename}", .{
					.typename = addable.typename(),
				});
	}

	///
	/// Attempts to pop the top-most value in `self`, pushing the unboxed value.
	///
	/// A [RuntimeError] is returned if `self` is out of memory or the top-most value is nil or not unboxable.
	///
	/// `self` is returned for function chaining.
	///
	pub fn boxed_get(self: *RuntimeEnv) RuntimeError!*RuntimeEnv {
		const unboxable = try self.expect_object(self.pop());

		defer self.release(unboxable);

		return switch (unboxable.internal().payload) {
			.boxed => |boxed| self.push(boxed),

			else => self.raise(error.TypeMismatch, "{typename} is not unboxable", .{
				.typename = unboxable.typename(),
			}),
		};
	}

	///
	/// Attempts to pop the top-most value in `self`, unboxing it, and replacing the contents with `value`.
	///
	/// A [RuntimeError] is returned if the top-most value is nil or not unboxable.
	///
	pub fn boxed_set(self: *RuntimeEnv, value: ?*RuntimeObj) RuntimeError!void {
		const unboxable = try self.expect_object(self.pop());

		defer self.release(unboxable);

		switch (unboxable.internal().payload) {
			.boxed => |*unboxed_value| {
				if (unboxed_value.*) |unboxed_object| {
					self.release(unboxed_object);
				}

				unboxed_value.* = if (value) |object| object.internal().acquire() else null;
			},

			else => return self.raise(error.TypeMismatch, "{typename} is not unboxable", .{
				.typename = unboxable.typename(),
			}),
		}
	}
};

///
/// Archetypes of errors that may occur in the virtual machine during run-time.
///
pub const RuntimeError = coral.io.AllocationError || error {
	IllegalState,
	TypeMismatch,
	BadOperation,
	BadSyntax,
};

///
/// Runtime-polymorphic data type for representing all data types supported by the virtual machine.
///
pub const RuntimeObj = opaque {
	const Internal = struct {
		ref_count: u16,

		payload: union (enum) {
			false,
			true,
			float: Float,
			fixed: Fixed,
			symbol: [*:0]const coral.io.Byte,
			vector2: Vector2,
			vector3: Vector3,
			syscall: *const Syscall,
			boxed: ?*RuntimeObj,

			string: struct {
				ptr: [*]coral.io.Byte,
				len: Fixed,

				const Self = @This();

				fn unpack(self: Self) []coral.io.Byte {
					coral.debug.assert(self.len >= 0);

					return self.ptr[0 .. @intCast(self.len)];
				}
			},

			dynamic: struct {
				ptr: [*]coral.io.Byte,

				const Self = @This();

				fn typeinfo(self: Self) *const Typeinfo {
					return @as(**const Typeinfo, @ptrCast(@alignCast(self.ptr))).*;
				}

				fn unpack(self: Self) []coral.io.Byte {
					return self.ptr[0 .. (@sizeOf(usize) + self.typeinfo().size)];
				}

				fn userdata(self: Self) []coral.io.Byte {
					const unpacked = self.unpack();
					const address_size = @sizeOf(usize);

					coral.debug.assert(unpacked.len >= address_size);

					return unpacked[address_size ..];
				}
			},
		},

		fn acquire(self: *Internal) *RuntimeObj {
			self.ref_count += 1;

			return @ptrCast(self);
		}
	};

	fn allocate(allocator: coral.io.Allocator, data: Internal) coral.io.AllocationError!*RuntimeObj {
		return @ptrCast(try coral.io.allocate_one(allocator, data));
	}

	fn internal(self: *const RuntimeObj) *Internal {
		return @constCast(@ptrCast(@alignCast(self)));
	}

	///
	/// Returns `true` if the value of `self` is equivalent to `other`, otherwise `false`.
	///
	pub fn equals(self: *RuntimeObj, other: *RuntimeObj) bool {
		return switch (self.internal().payload) {
			.false => other.internal().payload == .false,
			.true => other.internal().payload == .true,

			.fixed => |lhs_fixed| switch (other.internal().payload) {
				.fixed => |rhs_fixed| rhs_fixed == lhs_fixed,
				.float => |rhs_float| rhs_float == @as(Float, @floatFromInt(lhs_fixed)),
				else => false,
			},

			.float => |lhs_float| switch (other.internal().payload) {
				.float => |rhs_float| rhs_float == lhs_float,
				.fixed => |rhs_fixed| @as(Float, @floatFromInt(rhs_fixed)) == lhs_float,
				else => false,
			},

			.symbol => |lhs_symbol| switch (other.internal().payload) {
				.symbol => |rhs_symbol| lhs_symbol == rhs_symbol,
				else => false,
			},

			.boxed => |boxed| if (boxed) |object| self.equals(object) else false,

			.vector2 => |lhs_vector2| switch (other.internal().payload) {
				.vector2 => |rhs_vector2| lhs_vector2.equals(rhs_vector2),
				else => false,
			},

			.vector3 => |lhs_vector3| switch (other.internal().payload) {
				.vector3 => |rhs_vector3| lhs_vector3.equals(rhs_vector3),
				else => false,
			},

			.syscall => |lhs_syscall| switch (other.internal().payload) {
				.syscall => |rhs_syscall| lhs_syscall == rhs_syscall,
				else => false,
			},

			.string => |lhs_string| switch (other.internal().payload) {
				.string => |rhs_string| coral.io.are_equal(lhs_string.unpack(), rhs_string.unpack()),
				else => false,
			},

			.dynamic => |lhs_dynamic| switch (other.internal().payload) {
				.dynamic => |rhs_dynamic|
					lhs_dynamic.typeinfo() == rhs_dynamic.typeinfo() and
					lhs_dynamic.userdata().ptr == rhs_dynamic.userdata().ptr,

				else => false,
			},
		};
	}

	///
	/// Computes a hash for `self` based on the held type and value.
	///
	pub fn hash(self: *RuntimeObj) usize {
		return switch (self.internal().payload) {
			.false => 1237,
			.true => 1231,
			.float => |float| @bitCast(float),
			.fixed => |fixed| @intCast(@as(u32, @bitCast(fixed))),
			.symbol => |symbol| @intFromPtr(symbol),
			.vector2 => |vector| @bitCast(vector.to_scalars()),
			.vector3 => |vector| coral.io.jenkins_hash(@typeInfo(usize).Int, coral.io.bytes_of(&vector.to_scalars())),
			.syscall => |syscall| @intFromPtr(syscall),
			.boxed => |boxed| @intFromPtr(boxed),
			.string => |string| coral.io.djb2_hash(@typeInfo(usize).Int, string.unpack()),
			.dynamic => |dynamic| @intFromPtr(dynamic.typeinfo()) ^ @intFromPtr(dynamic.userdata().ptr),
		};
	}

	///
	/// Returns the name of the type held in `self`.
	///
	pub fn typename(self: *RuntimeObj) []const coral.io.Byte {
		return switch (self.internal().payload) {
			.false => "false",
			.true => "true",
			.float => "float",
			.fixed => "fixed",
			.symbol => "symbol",
			.vector2 => "vector2",
			.vector3 => "vector3",
			.syscall => "syscall",
			.boxed => "boxed",
			.string => "string",
			.dynamic => "dynamic",
		};
	}

	///
	/// Checks if `self` is a dynamic object with type information matching `typeinfo`, returning the userdata buffer if
	/// it is or `null` if it is not.
	///
	pub fn is_dynamic(self: *RuntimeObj, typeinfo: *const Typeinfo) ?[]coral.io.Byte {
		return switch (self.internal().payload) {
			.dynamic => |dynamic| if (dynamic.typeinfo() == typeinfo) dynamic.userdata() else null,
			else => null,
		};
	}

	///
	/// Checks if `self` is a false object (i.e. any object that evaluates to false), returning `true` if it is or
	/// `false` if it is not.
	///
	pub fn is_false(self: *RuntimeObj) bool {
		return !self.is_true();
	}

	///
	/// Checks if `self` is a fixed number object, returning the value if it is or `null` if it is not.
	///
	pub fn is_fixed(self: *RuntimeObj) ?Fixed {
		return switch (self.internal().payload) {
			.fixed => |fixed| fixed,
			else => null,
		};
	}

	///
	/// Checks if `self` is a numeric object (i.e. any object that behaves like a number), returning the supertype for
	/// further inspection if it is or `null` if it is not.
	///
	pub fn is_numeric(self: *RuntimeObj) ?Numeric {
		return switch (self.internal().payload) {
			.fixed => |fixed| .{.fixed = fixed},
			.float => |float| .{.float = float},
			.vector2 => |vector2| .{.vector2 = vector2},
			.vector3 => |vector3| .{.vector3 = vector3},
			else => null,
		};
	}

	///
	/// Checks if `self` is a string text object, returning the value if it is or `null` if it is not.
	///
	pub fn is_string(self: *RuntimeObj) ?[]const coral.io.Byte {
		return switch (self.internal().payload) {
			.string => |string| get: {
				coral.debug.assert(string.len > -1);

				break: get string.ptr[0 .. @intCast(string.len)];
			},

			.symbol => |symbol| coral.io.slice_sentineled(@as(u8, 0), symbol),
			else => null,
		};
	}

	///
	/// Checks if `self` is a symbo text object, returning the value if it is or `null` if it is not.
	///
	pub fn is_symbol(self: *RuntimeObj) ?[*:0]const coral.io.Byte {
		return switch (self.internal().payload) {
			.symbol => |symbol| symbol,
			else => null,
		};
	}

	///
	/// Checks if `self` is a true object (i.e. any object that evaluates to true), returning `true` if it is and
	/// `false` if it is not.
	///
	pub fn is_true(self: *RuntimeObj) bool {
		return switch (self.internal().payload) {
			.false => false,
			.true => true,
			.float => |float| float != 0,
			.fixed => |fixed| fixed != 0,
			.symbol => true,
			.vector2 => |vector| coral.io.all_equals(coral.io.bytes_of(&vector), 0),
			.vector3 => |vector| coral.io.all_equals(coral.io.bytes_of(&vector), 0),
			.syscall => true,
			.boxed => |boxed| boxed != null,
			.string => |string| string.len != 0,
			.dynamic => true,
		};
	}

	///
	/// Checks if `self` is a 2-component vector number object, returning the value if it is or `null` if it is not.
	///
	pub fn is_vector2(self: *RuntimeObj) ?Vector2 {
		return switch (self.internal().payload) {
			.vector2 => |vector2| vector2,
			else => null,
		};
	}

	///
	/// Checks if `self` is a 3-component vector number object, returning the value if it is or `null` if it is not.
	///
	pub fn is_vector3(self: *RuntimeObj) ?Vector3 {
		return switch (self.internal().payload) {
			.vector3 => |vector3| vector3,
			else => null,
		};
	}
};

///
/// Bridge function type between the virtual machine and native code.
///
pub const Syscall = fn (env: *RuntimeEnv) RuntimeError!?*RuntimeObj;

///
/// Type information for dynamically created object types
///
pub const Typeinfo = struct {
	name: []const coral.io.Byte,
	size: usize,
	destruct: ?*const fn (context: DestructContext) void = null,
	to_string: *const fn (context: ToStringContext) RuntimeError!*RuntimeObj = default_to_string,
	call: *const fn (context: CallContext) RuntimeError!?*RuntimeObj = default_call,
	get: *const fn (context: GetContext) RuntimeError!?*RuntimeObj = default_get,
	set: *const fn (context: SetContext) RuntimeError!void = default_set,

	///
	/// Context used for calling objects.
	///
	pub const CallContext = struct {
		env: *RuntimeEnv,
		userdata: []coral.io.Byte,
	};

	///
	/// Context used with objects that have special destruction semantics
	///
	pub const DestructContext = struct {
		env: *RuntimeEnv,
		userdata: []coral.io.Byte,
	};

	///
	/// Context used for getting indices of objects.
	///
	pub const GetContext = struct {
		env: *RuntimeEnv,
		userdata: []coral.io.Byte,

		///
		/// Pushes the index object of the get operation to the top of the stack.
		///
		/// *Note* the index is guaranteed to be a non-`null` value.
		///
		/// A [RuntimeError] is returned if the virtual machine is out of memory.
		///
		/// The [RuntimeEnv] is returned for function chaining.
		///
		pub fn push_index(self: *const GetContext) RuntimeError!*RuntimeEnv {
			coral.debug.assert(self.env.locals.values.len > 0);

			return self.env.push(self.env.locals.values[self.env.locals.values.len - 1]);
		}
	};

	///
	/// Context used for setting indices of objects to new values.
	///
	pub const SetContext = struct {
		env: *RuntimeEnv,
		userdata: []coral.io.Byte,

		///
		/// Pushes the index object of the set operation to the top of the stack.
		///
		/// *Note* the index is guaranteed to be a non-`null` value.
		///
		/// A [RuntimeError] is returned if the virtual machine is out of memory.
		///
		/// The [RuntimeEnv] is returned for function chaining.
		///
		pub fn push_index(self: *const SetContext) RuntimeError!*RuntimeEnv {
			coral.debug.assert(self.env.locals.values.len > 0);

			return self.env.push(self.env.locals.values[self.env.locals.values.len - 1]);
		}

		///
		/// Pushes the value of the set operation to the top of the stack.
		///
		/// A [RuntimeError] is returned if the virtual machine is out of memory.
		///
		/// The [RuntimeEnv] is returned for function chaining.
		///
		pub fn push_value(self: *const SetContext) RuntimeError!*RuntimeEnv {
			coral.debug.assert(self.env.locals.values.len > 1);

			return self.env.push(self.env.locals.values[self.env.locals.values.len - 2]);
		}
	};

	///
	/// Context used for converting objects to string representations.
	///
	pub const ToStringContext = struct {
		env: *RuntimeEnv,
		userdata: []coral.io.Byte,
	};

	fn default_call(context: CallContext) RuntimeError!?*RuntimeObj {
		return context.env.raise(error.BadOperation, "this dynamic object is not callable", .{});
	}

	fn default_get(context: GetContext) RuntimeError!?*RuntimeObj {
		return context.env.raise(error.BadOperation, "this dynamic object is not get-indexable", .{});
	}

	fn default_set(context: SetContext) RuntimeError!void {
		return context.env.raise(error.BadOperation, "this dynamic object is not set-indexable", .{});
	}

	fn default_to_string(context: ToStringContext) RuntimeError!*RuntimeObj {
		return (try context.env.new_string("{}")).pop().?;
	}
};

///
/// 2-component vector type.
///
pub const Vector2 = gfx.lina.Vector2;

///
/// 3-component vector type.
///
pub const Vector3 = gfx.lina.Vector3;

///
/// Higher-level wrapper for [RuntimeEnv.index_get] that makes it easier to index [Fixed] keys of objects.
///
pub fn get_at(env: *RuntimeEnv, indexable: *RuntimeObj, index: Fixed) RuntimeError!?*RuntimeObj {
	const index_number = (try env.new_fixed(index)).pop();

	defer env.discard(index_number);

	return (try (try env.push(indexable)).index_get(index_number)).pop();
}

///
/// Higher-level wrapper for [RuntimeEnv.index_get] that makes it easier to index field keys of objects.
///
pub fn get_field(env: *RuntimeEnv, indexable: *RuntimeObj, field: []const coral.io.Byte) RuntimeError!?*RuntimeObj {
	const field_symbol = (try env.new_symbol(field)).pop().?;

	defer env.release(field_symbol);

	return (try (try env.push(indexable)).index_get(field_symbol)).pop();
}

///
/// Higher-level wrapper for [RuntimeEnv.index_get] that makes it easier to index string keys of objects.
///
pub fn get_key(env: *RuntimeEnv, indexable: *RuntimeObj, key: []const coral.io.Byte) RuntimeError!?*RuntimeObj {
	const key_string = (try env.new_string(key)).pop();

	defer env.discard(key_string);

	return (try (try env.push(indexable)).index_get(key_string)).pop();
}
