const coral = @import("coral");

const kym = @import("../kym.zig");

associative: RefTable,
contiguous: RefList,

const RefList = coral.list.Stack(?*kym.RuntimeObj);

const RefTable = coral.map.Table(*kym.RuntimeObj, *kym.RuntimeObj, struct {
	pub const hash = kym.RuntimeObj.hash;

	pub const equals = kym.RuntimeObj.equals;
});

const Self = @This();

pub fn deinit(self: *Self, env: *kym.RuntimeEnv) void {
	{
		var field_iterable = self.associative.as_iterable();

		while (field_iterable.next()) |entry| {
			env.release(entry.key);
			env.release(entry.value);
		}
	}

	self.associative.deinit();

	while (self.contiguous.pop()) |value| {
		if (value) |ref| {
			env.release(ref);
		}
	}

	self.contiguous.deinit();
}

pub fn init(env: *kym.RuntimeEnv) Self {
	return .{
		.associative = RefTable.init(env.allocator, .{}),
		.contiguous = RefList.init(env.allocator),
	};
}

pub const typeinfo = &kym.Typeinfo{
	.size = @sizeOf(Self),
	.name = "table",
	.destruct = typeinfo_destruct,
	.get = typeinfo_get,
	.set = typeinfo_set,
};

fn typeinfo_destruct(context: kym.Typeinfo.DestructContext) void {
	@as(*Self, @ptrCast(@alignCast(context.userdata))).deinit(context.env);
}

fn typeinfo_get(context: kym.Typeinfo.GetContext) kym.RuntimeError!?*kym.RuntimeObj {
	const table = @as(*Self, @ptrCast(@alignCast(context.userdata)));
	const index = (try context.push_index()).pop().?;

	defer context.env.release(index);

	if (index.is_fixed()) |fixed| {
		if (fixed < 0) {
			// TODO: Negative indexing.
			unreachable;
		}

		if (fixed < table.contiguous.values.len) {
			return table.contiguous.values[@intCast(fixed)];
		}
	}

	if (table.associative.lookup(index)) |value| {
		return value;
	}

	return null;
}

fn typeinfo_set(context: kym.Typeinfo.SetContext) kym.RuntimeError!void {
	const table = @as(*Self, @ptrCast(@alignCast(context.userdata)));
	const index = (try context.push_index()).pop().?;

	errdefer context.env.release(index);

	if (index.is_fixed()) |fixed| {
		if (fixed < 0) {
			// TODO: Negative indexing.
			unreachable;
		}

		if (fixed < table.contiguous.values.len) {
			const maybe_replacing = &table.contiguous.values[@intCast(fixed)];

			if (maybe_replacing.*) |replacing| {
				context.env.release(replacing);
			}

			if ((try context.push_value()).pop()) |value| {
				errdefer context.env.release(value);

				maybe_replacing.* = value;
			} else {
				maybe_replacing.* = null;
			}

			return;
		}
	}

	const value = (try context.push_value()).pop() orelse {
		if (table.associative.remove(index)) |removed| {
			context.env.release(removed.key);
			context.env.release(removed.value);
		}

		return;
	};

	errdefer context.env.release(value);

	if (try table.associative.replace(index, value)) |replaced| {
		context.env.release(replaced.key);
		context.env.release(replaced.value);
	}
}
